/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ132MvMicKS
 *
 * de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ132MvMicKS is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ132MvMicKS is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ132MvMicKS.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ132MvMicKS.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ132MvMicKS ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.mv.Fg003Typ132MvMicKS wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7.conversion.mv;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.kex.tls.osi7.*;
import de.bsvrz.kex.tls.osi7.conversion.Fg003Default;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Umsetzungsklasse für FG 3, Typ 132, Firma MicKS in Mecklenburg Vorpommern Glättewarnung
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */

public class Fg003Typ132MvMicKS extends Fg003Default {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/**
	 * DebugLogger für Debug-Ausgaben
	 */
	private static final Debug debug = Debug.getLogger();
// Private ----------------------------------------------------------------
	/**
	 * Tabelle, in der das Verhalten beim Empfang von Attributgruppen eingetragen wird.
	 */
	private final Object[][] davToTls = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp

	};

	/**
	 * Tabelle, in der das Verhalten beim Empfang von TLS De-Blöcken eingetragen wird.
	 */
	private final Object[][] tlsToDav = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp
			{"atg.tlsUfdErgebnisMeldungFahrbahnGlätteFBGMicKS", "asp.tlsAntwort", "3", "132", "4", "typ.deUfd"},
	};

// ------------------------------------------------------------------------
// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
// ------------------------------------------------------------------------

	/**
	 * Objekterzeugung mit Anmeldung der globalen und lokalen Umsetzungseinträge
	 *
	 * @param daf DaV
	 * @param eak Eak
	 * @param de De
	 */
	public Fg003Typ132MvMicKS(ClientDavInterface daf, Eak eak, De de) {
		super(daf, eak, de);
		subscribeAllData(tlsToDav, davToTls);
	}

// ------------------------------------------------------------------------
// SONSTIGE METHODEN
// ------------------------------------------------------------------------

	/**
	 * FG 3 : Typ: 132 Id: 4 Name: Fahrbahnglätte (Antwort) TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks [ 3 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 1..254 ]
	 * Byte 3           Typ der DE-Daten                        [ 132 ]
	 * Byte 4           ZeitreserveReifGlätte                   siehe unten
	 * Byte 5           ZeitreserveEisGlätte                    siehe unten
	 * Byte 5           AlarmGlattEis                           siehe unten
	 * </pre>
	 * Inhalt Byte: Fahrbahnglätte:
	 * <pre>
	 *        0:   keine Glätte vorhersehbar
	 *        3:   Glättegefahr durch Nässe
	 *       65:   Glättegefahr durch Schnee/Schneematsch
	 *       66:   Glättegefahr durch Eis/Glatteis
	 *       67:   Glättegefahr durch Rauhreif
	 *       68:   Glättegefahr durch unterkühlte Fahrbahn
	 *      254:   keine Glätte vorhersehbar
	 *      255:   nicht ermittelbar
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg003Typ132Id004AtgTlsUfdErgebnisMeldungFahrbahnGlätteFBGMicKSAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// _Messwerte_ aus Telegramm -----------------------------------------------------------------------------------
		int _Fahrbahnglätte_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("Fahrbahnglätte").set(_Fahrbahnglätte_);

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}
}
