/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne.Fg254Duerr
 *
 * de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne.Fg254Duerr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne.Fg254Duerr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne.Fg254Duerr.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne.Fg254Duerr.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne.Fg254Duerr ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne.Fg254Duerr wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7.conversion.nw.uz.loehne;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.IntegerValueState;
import de.bsvrz.kex.tls.osi7.*;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.Calendar;
import java.util.GregorianCalendar;

/**
 * Umsetungsmodul für die FG 254, Standard (alle Typen der FG 254 gemäß TLS 2002) Wird standardmäßig für alle FG 254 De
 * verwendet, bei denen keine spezielle Umsezungsklasse in der Konfiguration angegeben wurde.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */

public class Fg254Duerr extends FgAllDuerr {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/** DebugLogger für Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();
// Private ----------------------------------------------------------------
	/**
	 * Tabelle, in der das Verhalten beim Empfang von Attributgruppen eingetragen wird. Jede Tabellenzeile hat folgenden
	 * Aufbau:
	 * <pre>
	 * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
	 * z.B.
	 * "atg.tlsGloKanalSteuerung", "asp.parameterSoll", "1", "29", "2", "typ.deLve"
	 * </pre>
	 * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
	 * Empfangsanmeldung als Empfänger.</li> <li> In der Methode {@link #update} wird für jede zur
	 * Attributgruppen/Aspket-Kombination passende Kombination eine Methode aufgerufen, deren Namen aus dem Tabelleneintrag
	 * wie folgt gebildet wird (hier für den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspParameterSoll(ResultData)</code><br>
	 * Dabei enthält <code>ResultData</code> die Daten der zu verarbeitenden ATG/ASP-Kombination.</li> <li> Für jede
	 * Attributguppen/Aspekt Kombination erfolgt eine Empfangsanmeldung als Empfänger. Sind mehrere passende Eintrage
	 * vorhanden, werden entsprechend mehrere Methodenaufrufe durchgeführt. Dadurch können mehrere ATG/ASP-Kombinationen in
	 * einen De-Block abgebildet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eines Fachmodells):</li>
	 * <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
	 */
	private final Object[][] davToTls = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp
			{"atg.tlsGloDeBlockAbruf", "asp.tlsAbruf", "255", "255", "255", "typ.deSys"},
			{"atg.tlsGloDeBlockSenden", "asp.tlsVorgabe", "255", "255", "255", "typ.deSys"},
			{"atg.tlsGloKanalSteuerung", "asp.parameterSoll", "255", "29", "2", "typ.deSys"},
			{"atg.tlsGloGeographischeKenndaten", "asp.tlsVorgabe", "255", "36", "3", "typ.deSys"},
			{"atg.tlsSysZeitSynchronisation", "asp.tlsVorgabe", "254", "18", "2", "typ.deSys"},
			{"atg.tlsSysStatischeGeräteKennDaten", "asp.tlsVorgabe", "254", "32", "3", "typ.deSys"},
			{"atg.tlsSysKonfigurationsTabelle", "asp.tlsVorgabe", "254", "34", "3", "typ.deSys"},
			{"atg.tlsSysOSI3RoutingFeld", "asp.tlsVorgabe", "254", "35", "3", "typ.deSys"},
			{"atg.tlsSysKnotenNummer", "asp.tlsVorgabe", "254", "37", "3", "typ.deSys"},
			{"atg.tlsSysReset", "asp.tlsVorgabe", "254", "38", "2", "typ.deSys"}
	};

	/**
	 * Tabelle, in der das Verhalten beim Empfang von TLS De-Blöcken eingetragen wird. Jede Tabellenzeile hat folgenden
	 * Aufbau:
	 * <pre>
	 * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
	 * z.B.
	 * "atg.tlsGloKanalSteuerung", "asp.tlsAntwort", "1", "29", "2", "typ.deLve"
	 * </pre>
	 * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
	 * Sendeanmeldung als Quelle.</li> <li> In der Methode {@link #processDeBlock} wird für jede zum De-Block passende
	 * Fg/Id/Typ-Kombination eine Methode aufgerufen, deren Namen aus dem Tabelleneintrag wie folgt gebildet wird (hier für
	 * den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspTlsAntwort(DeBlock, Data,
	 * DataDescription)</code><br> Dabei enthält der <code>deBlock</code> die Daten des zu verarbeitenden De-Blocks und
	 * <code>data</code> ein Datenobjekt der passenden Attributgruppen/Aspekt-Kombination, welchem direkt die Inhalte des
	 * De-Blocks zugewiesen werden können.<code>DataDescription</code> enthält die passende {@link
	 * DataDescription} für die zu sendenden Daten.</li> <li> Für jede Attributguppen/Aspekt
	 * Kombination erfolgt eine Sendeanmeldung als Quelle. Sind mehrere passende Eintrage vorhanden, werden entsprechend
	 * mehrere Methodenaufrufe durchgeführt. Dadurch kann ein De-Block mehrfach ausgewert und an verschiedene
	 * Attributgruppen weitergerleitet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eine
	 * Fachmodells):</li> <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
	 */
	private final Object[][] tlsToDav = new Object[][]{
			// Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp
			{"atg.tlsGloDeFehler", "asp.tlsAntwort", "254", "1", "1", "typ.deSys"},
			{"atg.tlsGloKanalSteuerung", "asp.parameterIst", "255", "29", "2", "typ.deSys"},
			{"atg.tlsGloGeographischeKenndaten", "asp.tlsZustand", "255", "36", "3", "typ.deSys"},
			{"atg.tlsSysErgänzendeDeFehlerMeldung", "asp.tlsAntwort", "254", "14", "1", "typ.deSys"},
			{"atg.tlsSysNegativeQuittung", "asp.tlsAntwort", "254", "16", "2", "typ.deSys"},
			{"atg.tlsSysInitialisierungsMeldung", "asp.tlsAntwort", "254", "17", "2", "typ.deSys"},
			{"atg.tlsSysKommunikationsStatus", "asp.tlsAntwort", "254", "19", "2", "typ.deSys"},
			{"atg.tlsSysPositiveQuittung", "asp.tlsAntwort", "254", "28", "3", "typ.deSys"},
			{"atg.tlsSysStatischeGeräteKennDaten", "asp.tlsZustand", "254", "32", "3", "typ.deSys"},
			{"atg.tlsSysKonfigurationsTabelle", "asp.tlsZustand", "254", "34", "3", "typ.deSys"},
			{"atg.tlsSysOSI3RoutingFeld", "asp.tlsZustand", "254", "35", "3", "typ.deSys"},
			{"atg.tlsSysKnotenNummer", "asp.tlsZustand", "254", "37", "3", "typ.deSys"},

	};

// ------------------------------------------------------------------------
// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
// ------------------------------------------------------------------------

	/**
	 * Erzeugt ein Objekt vom Typ Fg001Default
	 *
	 * @param daf DaV
	 * @param eak Eak
	 * @param de De
	 */
	public Fg254Duerr(ClientDavInterface daf, Eak eak, De de) {
		super(daf, eak, de);
		subscribeAllData(tlsToDav, davToTls);
	}

// ------------------------------------------------------------------------
// SONSTIGE METHODEN
// ------------------------------------------------------------------------

	/**
	 * FG 254 : Typ: 1 Id: 1 Name: DE-Fehler (Antwort). Verwendet globale Behandlung De-Fehler in {@link
	 * #Fg255Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort}.
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = super.Fg255Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort(deBlock, data, dataDescription);

		// Über die FG 254 werden gestörte SM oder EAK gemeldet. Für diese werden alle angemeldeten Daten an den
		// zugeordneten DE auf "keine Daten" gesetzt, da deren Zustand unbekannt ist.
		IntegerValueState DEFehlerStatus = data.getUnscaledValue("DEFehlerStatus").getState();
		if (DEFehlerStatus != null) {
			debug.fine("DEFehlerStatus: " + DEFehlerStatus.getName());
			if (DEFehlerStatus.getName().equals("StörSM")) {
				if (_de.getDeChannel() == 0) {// SM gestört, Knotennummer ermittlen.
					int[] osi7Adress = new int[]{deBlock.getOsi7Adress()};
					TlsApplicationLayer.writeNoDataToUnreachableDe(osi7Adress, _eakConversionModul.getTimeStamp());
				}
				else {// EAK gestört
					Sm sm = _eak.getSm();
					Eak eak = sm.getEak(_de.getDeChannel());// Eak mit der gesuchten Osi2Adresse am Sm ermitteln.
					if (eak != null) {
						TlsApplicationLayer.writeNoDataToUnreachableDe(eak, _eakConversionModul.getTimeStamp());
					}
				}
			}
		}
		return resultData;
	}

	/**
	 * FG 254 : Typ: 14 Id: 1 Name: Ergänzende DE-Fehlermeldung (Antwort) TLS-Datenbeschreibung:
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          [ 9..29 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 1..254 ]
	 * Byte 3           Typ der DE-Daten                        [ 14 ]
	 * Byte 4           Herstellercode                          (s. Anh. 7, Kap. 2)
	 * Byte 5           Anzahl TLS - Fehlerbytes                [ 4 ]
	 * Byte 6           TLS - Fehlerbyte 1                      Codierung siehe unten
	 * Byte 7           TLS - Fehlerbyte 2                      Codierung siehe unten
	 * Byte 8           TLS - Fehlerbyte 3                      Codierung siehe unten
	 * Byte 9           TLS - Fehlerbyte 4                      Codierung siehe unten
	 * Byte 10          Anzahl Hersteller Fehlerbytes           [ 0..20 ]
	 * Byte 11          Hersteller-Fehlerbyte 1                 herstellerspez. Codierung
	 * Byte..           Hersteller-Fehlerbyte n                 herstellerspez. Codierung
	 * </pre>
	 * <br> Inhalt TLS-Fehlerbytes:<br> Die Fehlerbytes sind bitweise codiert. Jedes Bit definiert einen bestimmten Fehler,
	 * wobei Bit = 0 ok und Bit = 1 Fehler bedeutet. Es können auch mehrere Bits gleichzeitig gesetzt werden. Der Inhalt
	 * der Fehlerbytes ist optional. D.h. bei einer Realisierung dieses Telegrammtyps müssen vom Hersteller nicht alle in
	 * der Tabelle definierten Fehlercodes erfüllt werden. Für spätere Definitionen reservierte Bits bzw. nicht verwendete
	 * Bits sind gleich Null zu setzen.:<br>
	 * <pre>
	 *         Bit 7      Bit 6       Bit 5       Bit 4          Bit 3         Bit 2       Bit 1       Bit 0
	 * Byte 1  Reserve    Reserve     RNR am      Lokalbus       Reserve       Reserve     Reserve     Teilstörung
	 *                                Lokalbus1   ausgefallen
	 * Byte 2  Reserve    Reserve     Reserve     Reserve        Reserve       Reserve     Reserve     Fehler DE-
	 *                                                                                                 Zuordnung2
	 * Byte 3  Reserve    Reserve     Reserve     Reserve        Reserve       Reserve     Reserve     Reserve
	 * Byte 4  Reserve    Reserve     Reserve     Reserve        Reserve       Reserve     Reserve     Reserve
	 * 1) nur in Verbindung mit Stö/SM=1 im Byte 4 der DE-Block-Struktur im Typ 1
	 * 2 nur in Verbindung mit Proj-F=1
	 * </pre>
	 * Siehe auch Anhang 6, Teil 1, Kapitel 3.3.1.6.
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ014Id001AtgTlsSysErgänzendeDeFehlerMeldungAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Herstellercode-Byte aus Telegramm ---------------------------------------------------------------------------
		int _Herstellercode_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("Hersteller").set(_Herstellercode_);

		// Fehler-Bytes aus Telegramm -----------------------------------------------------------------------------------
		deBlock.getDeData().readUnsignedByteAsInt();// Anzahl der folgenden Bytes überlesen, da immer == 4 !
		int _TlsFehler1_ = deBlock.getDeData().readUnsignedByteAsInt();
		int _TlsFehler2_ = deBlock.getDeData().readUnsignedByteAsInt();
		int _TlsFehler3_ = deBlock.getDeData().readUnsignedByteAsInt();
		int _TlsFehler4_ = deBlock.getDeData().readUnsignedByteAsInt();

		Data.NumberArray _TlsFehler_ = data.getItem("FehlercodesTLS").getUnscaledArray("Fehler");
		_TlsFehler_.setLength(4);
		_TlsFehler_.getValue(0).set(_TlsFehler1_);
		_TlsFehler_.getValue(1).set(_TlsFehler2_);
		_TlsFehler_.getValue(2).set(_TlsFehler3_);
		_TlsFehler_.getValue(3).set(_TlsFehler4_);

		// Hersteller Fehler-Bytes aus Telegramm -----------------------------------------------------------------------
		int count = deBlock.getDeData().readUnsignedByteAsInt();
		int[] _HerstellerFehler_ = new int[count];
		for (int i = 0; i < count; i++) {
			_HerstellerFehler_[i] = deBlock.getDeData().readUnsignedByteAsInt();
		}

		Data.NumberArray _Fehler_ = data.getItem("FehlercodesHersteller").getUnscaledArray("Fehler");
		_Fehler_.setLength(count);
		for (int i = 0; i < count; i++) {
			_Fehler_.getValue(i).set(_HerstellerFehler_[i]);
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 254 : Typ: 16 Id: 2 Name: Negative Quittung (Antwort) TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks [ 4 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 bzw. entspricht OSI2-Adresse ]
	 * Byte 3           Typ der DE-Daten                        [ 16 ]
	 * Byte 4           Fehlerursache                           siehe unten
	 * Byte 5           Herstellercode                          siehe Liste in Anhang 7, Kap. 2
	 * </pre>
	 * Inhalt Byte: Fehlerursache unter FG 254:
	 * <pre>
	 *        0:   sonstige Fehlerursache
	 *        1:   unbekannte oder nicht auswertbare ID
	 *        2:   unbekannter oder nicht auswertbarer Typ
	 *        3:   Kein Rückmeldung möglich, da Kanal gestört
	 *        4:   frei
	 *        5:   Kommunikationsstatus enthält ungültigen Wert
	 *        6:   Nachricht DE-Zuordnung enthält mehr E/A-Kanäle als phys. vorhanden
	 *        7:   Nachricht DE-Zuordnung enthält unzulässige DE mit Sammeladresse (???1 1111b)
	 *        8:   Nachricht DE-Zuordnung enthält zu einem E/A-Kanal die falsche FG
	 *    9..63:   noch nicht definiert
	 * 64 ..127:   Adress- und Strukturfehler
	 *       64:   nicht vorhandene Funktionsgruppe
	 *       65:   nicht vorhandener DE
	 *       66:   Richtungsbit im Abruftelegramm war "1"
	 *       67:   Jobnummer in Abruftelegramm ist 0
	 *       68:   Anzahl Einzeltelegramme ist nicht mit gesamter OSI7-Länge vereinbar
	 *       69:   Längenbyte des Einzeltelegramms ist nicht mit gesamter OSI7-Länge vereinbar
	 *       70:   Anzahl DE-Blöcke ist nicht mit gesamter OSI7-Länge vereinbar
	 *       71:   Länge DE-Block ist nicht mit gesamter OSI7-Länge vereinbar
	 *       72:   Anzahl DE-Blöcke ist nicht mit dem Längenbyte des Einzeltelegramms vereinbar
	 *       73:   Länge DE-Block ist nicht mit dem Längenbyte des Einzeltelegramms vereinbar
	 *       74:   Empfänger oder Übertragungsweg ist gestört
	 *       75:   Anzahl der DE-Blöcke ist unzulässig
	 *       76:   falsche Knotennummer
	 *       77:   fehlerhafte OSI-3 Routinginformation
	 *       78:   Anzahl Einzeltelegramme unzulässig
	 *  79..127:   noch nicht definiert
	 * 128..255:   frei für Herstellerdefinitionen
	 * </pre>
	 * Inhalt Byte Herstellercode:<br> Der Herstellercode ist ein für jeden Hersteller eindeutig zugeordneter Code, der der
	 * Zentrale die Zuordnung der Nachrichten erleichtern soll.
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ016Id002AtgTlsSysNegativeQuittungAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		return super.Fg255Typ016Id002AtgTlsGloNegativeQuittungAspTlsAntwort(deBlock, data, dataDescription);
	}

	/**
	 * FG 254 : Typ: 17 Id: 2  Name: Initialisierungsmeldung (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks [ 2 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0, bzw. entspricht OSI2 ]
	 * Byte 3           Typ der DE-Daten                        [ 17 ]
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ017Id002AtgTlsSysInitialisierungsMeldungAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Datensatz enthält keine Daten, es wird ein konstanter Datensatz erzeugt, dessen Zeitstempel maßgebend ist.
		data.getUnscaledValue("InitialisierungsMeldung").set(0);

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 254 : Typ: 18 Id: 2  Name: Zeitsynchronisation (Abruf). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks [ 9 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [0..254, 255]
	 * Byte 3           Typ der DE-Daten                        [ 18 ]
	 * Byte 4           Zeitbereich / Stunde                    [ 0/1 * 128 + 0..23 ]
	 * Byte 5           Minute                                  [ 0..59 ]
	 * Byte 6           Sekunde                                 [ 0..59 ]
	 * Byte 7           Tag                                     [ 1..31 ]
	 * Byte 8           Monat                                   [ 1..12 ]
	 * Byte 9           Jahr                                    [ 0..99 ]
	 * Byte 10          Wochentag                               [ Mo..So = 1..7 ]
	 * </pre>
	 * Inhalt Byte Zeitbereich/Stunde<br>
	 * <pre>
	 * Bit 0...6:   Angabe der Stunde, binär codiert
	 * Bit 7 = 0:   Normalzeit (MEZ)
	 * Bit 7 = 1:   Sommerzeit (MESZ)
	 * </pre>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg254Typ018Id002AtgTlsSysZeitSynchronisationAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		long _Zeitpunkt_;
		int _Jahr_;
		int _Monat_;
		int _Tag_;
		int _Wochentag_;
		int _Stunde_;
		int _Minute_;
		int _Sekunde_;

		// Umsetzung
		Data data = result.getData();// Datensatz holen

		_Zeitpunkt_ = data.getTimeValue("Zeitpunkt").getMillis();

		GregorianCalendar gregorianCalendar = new GregorianCalendar();
		gregorianCalendar.setTimeInMillis(_Zeitpunkt_);

		_Jahr_ = (gregorianCalendar.get(Calendar.YEAR) - 2000);
		_Monat_ = (gregorianCalendar.get(Calendar.MONTH) + 1);
		_Tag_ = gregorianCalendar.get(Calendar.DAY_OF_MONTH);
		_Wochentag_ = gregorianCalendar.get(Calendar.DAY_OF_WEEK) - 1;// TLS Mo=1, Kalender Sonntag=1;
		if (_Wochentag_ == 0) _Wochentag_ = 7;// Sonderbehandlung für den Sonntag

		_Stunde_ = gregorianCalendar.get(Calendar.HOUR_OF_DAY);
		if (gregorianCalendar.get(Calendar.DST_OFFSET) > 0) {
			_Stunde_ += 128;
		}
		_Minute_ = gregorianCalendar.get(Calendar.MINUTE);
		_Sekunde_ = gregorianCalendar.get(Calendar.SECOND);

		// DeBlock als TlsData aufbauen...
		TlsData deApplicationData = new TlsData(7);
		deApplicationData.writeIntAsByte(_Stunde_);
		deApplicationData.writeIntAsByte(_Minute_);
		deApplicationData.writeIntAsByte(_Sekunde_);
		deApplicationData.writeIntAsByte(_Tag_);
		deApplicationData.writeIntAsByte(_Monat_);
		deApplicationData.writeIntAsByte(_Jahr_);
		deApplicationData.writeIntAsByte(_Wochentag_);

		// DeBlock erzeugen
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 18, 2, _de.getDeChannel());
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}

	/**
	 * FG254 : Typ: 28 Id: 3  Name: Positive Quittung (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks [ 2 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [0 bzw. entspricht OSI2-Adresse ]
	 * Byte 3           Typ der DE-Daten                        [ 28 ]
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ028Id003AtgTlsSysPositiveQuittungAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Versuch, zugehörigen (Parameter)DeBlock zu finden, falls vorhanden entsprechende Methode so aufrufen, als
		// wäre dieser DeBlock als Antwort "gespiegelt" worden, ...
		DeBlock sendDeBlock = TlsApplicationLayer.getSendDeBlock(_eakConversionModul.getJobNumber());
		if (sendDeBlock != null) {
			debug.warning("Positive Quittung weitergeleitet:\n" + sendDeBlock);
			_de.getConversionModul().processDeBlock(sendDeBlock);
			data.getTextValue("DeBlock").setText(sendDeBlock.toString());
		}
		else {
			// ...ansonsten  Warnung ausgeben.
			data.getTextValue("DeBlock").setText("Fehler: DeBlock für diese Quittung (Jobnummer = "
			                                     + _eakConversionModul.getJobNumber() +
			                                     ") konnte nicht ermittelt werden!");
			debug.warning("Positive Quittung zum DeBlock konnte nicht ermittelt werden\n" + deBlock);
			//Ergebnisdatensatz erzeugen...
			resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		}
		return resultData;
	}

	/**
	 * FG 254 : Typ: 32 Id: 3 Name: Statische Gerätekenndaten (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 bzw. entspricht OSI2-Adresse ]
	 * Byte 3           Typ der DE-Daten                        [ 32 ]
	 * Byte 4           Hersteller Herstellercode               (s. Anh. 7, Kap. 2)
	 * Byte 5           Gerätekennungsbyte 1                    vom Hersteller vergeben
	 * Byte..           Gerätekennungsbyte n
	 * </pre>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg254Typ032Id003AtgTlsSysStatischeGeräteKennDatenAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _Herstellercode_;
		int _Kenndaten_;

		// Umsetzung
		Data data = result.getData();// Datensatz holen

		int count = data.getArray("Kenndaten").getLength();
		TlsData deApplicationData = new TlsData(1 + count);

		_Herstellercode_ = data.getUnscaledValue("Hersteller").intValue();
		deApplicationData.writeIntAsByte(_Herstellercode_);

		for (int i = 0; i < count; i++) {
			_Kenndaten_ = data.getArray("Kenndaten").getItem(i).asUnscaledValue().intValue();
			deApplicationData.writeIntAsByte(_Kenndaten_);
		}

		// DeBlock erzeugen
		try {
			// Statische Gerätekenndaten sind 'eigentlich' nicht in SendeRichtung definiert
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 32, 2, _de.getDeChannel());
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}

	/**
	 * FG 254 : Typ: 32 Id: 3 Name: Statische Gerätekenndaten (Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 bzw. entspricht OSI2-Adresse ]
	 * Byte 3           Typ der DE-Daten                        [ 32 ]
	 * Byte 4           Hersteller Herstellercode               (s. Anh. 7, Kap. 2)
	 * Byte 5           Gerätekennungsbyte 1                    vom Hersteller vergeben
	 * Byte..           Gerätekennungsbyte n
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ032Id003AtgTlsSysStatischeGeräteKennDatenAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Herstellercode-Byte aus Telegramm ---------------------------------------------------------------------------
		int _Herstellercode_ = deBlock.getDeData().readUnsignedByteAsInt();
		data.getUnscaledValue("Hersteller").set(_Herstellercode_);

		// Kenndaten-Byte aus Telegramm --------____--------------------------------------------------------------------
		int count = deBlock.getDeData().length() - deBlock.getDeData().lastReadPosition() - 1;
		int _Kenndaten_;
		data.getArray("Kenndaten").setLength(count);
		for (int i = 0; i < count; i++) {
			_Kenndaten_ = deBlock.getDeData().readUnsignedByteAsInt();
			data.getArray("Kenndaten").getItem(i).asUnscaledValue().set(_Kenndaten_);
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 254 : Typ: 34 Id: 3 Name: Konfigurationstabelle (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 = Systemsteuerung des SM ]
	 * Byte 3           Typ der DE-Daten                        [ 34 ]
	 * Byte 4           Anzahl Tabellenzeilen                   [ 0, 1..56]
	 * Byte 5           Zeile 1: E/A-Konzentrator               OSI-2-Adresse
	 * Byte 6           Zeile 1: phys. E/A-Kanal-Nr.            Fortlaufende, phys. Nummer des Kanals
	 * Byte 7           Zeile 1: FG                             Funktionsgruppe des Kanals
	 * Byte 8           Zeile 1: DE                             Zuzuweisende DE
	 * Byte ..          Zeile n: E/A-Konzentrator               OSI-2-Adresse
	 * Byte ..          Zeile n: phys. E/A-Kanal-Nr.            Fortlaufende, phys. Nummer des Kanals
	 * Byte ..          Zeile n: FG                             Funktionsgruppe des Kanals
	 * Byte ..          Zeile n: DE                             Zuzuweisende DE
	 * </pre>
	 * Inhalt Byte: Anzahl Tabellenzeilen:<br> Gibt die Länge der Konfiguration/Soll-Tabelle in Zeilen an. Die Gesamtlänge
	 * der Tabelle ergibt sich aus der Multiplikation mit der festen Zeilenlänge von 4 Bytes. Die Angabe 0 löscht die
	 * vorhandene Konfigurationstabelle.<br> <br> Inhalt Byte: E/A-Konzentrator:<br> In diesem Byte ist die OSI-2-Adresse
	 * abgelegt, über die der E/A-Konzentrator am Lokalbus adressiert wird. Ist der E/A-Konzentrator intern im SM
	 * erreichbar, so wird in diesem Byte 0 oder ein Wert größer 31 (herstellerabhängig) eingetragen.<br> <br> Inhalt Byte:
	 * E/A-Kanal-Nr.:<br> Die interne, phys. E/A-Kanalnummer im E/A-Konzentrator.<br> <br> Inhalt Byte: FG:<br> Die FG gibt
	 * die Funktionsgruppe an, zu der der Kanal gehört.<br> <br> Inhalt Byte: DE:<br> Der DE gibt die logische Kanalnummer
	 * an, mit welcher die Adressierung dieses Kanals vorgenommen werden soll.<br>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg254Typ034Id003AtgTlsSysKonfigurationsTabelleAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _OSI2Adresse_;
		int _Kanal_;
		int _FG_;
		int _DE_;

		// Umsetzung
		Data data = result.getData();// Datensatz holen

		int count = data.getArray("KonfigurationsEintrag").getLength();
		TlsData deApplicationData = new TlsData(1 + (count * 4));
		deApplicationData.writeIntAsByte(count);

		for (int i = 0; i < count; i++) {
			_OSI2Adresse_ = data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("OSI2Adresse").intValue();
			_Kanal_ = data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("Kanal").intValue();
			_FG_ = data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("FG").intValue();
			_DE_ = data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("DE").intValue();
			deApplicationData.writeIntAsByte(_OSI2Adresse_);
			deApplicationData.writeIntAsByte(_Kanal_);
			deApplicationData.writeIntAsByte(_FG_);
			deApplicationData.writeIntAsByte(_DE_);
		}

		// DeBlock erzeugen
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 34, 3, _de.getDeChannel());
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}

	/**
	 * FG 254 : Typ: 34 Id: 3 Name: Konfigurationstabelle (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 = Systemsteuerung des SM ]
	 * Byte 3           Typ der DE-Daten                        [ 34 ]
	 * Byte 4           Anzahl Tabellenzeilen                   [ 0, 1..56]
	 * Byte 5           Zeile 1: E/A-Konzentrator               OSI-2-Adresse
	 * Byte 6           Zeile 1: phys. E/A-Kanal-Nr.            Fortlaufende, phys. Nummer des Kanals
	 * Byte 7           Zeile 1: FG                             Funktionsgruppe des Kanals
	 * Byte 8           Zeile 1: DE                             Zuzuweisende DE
	 * Byte ..          Zeile n: E/A-Konzentrator               OSI-2-Adresse
	 * Byte ..          Zeile n: phys. E/A-Kanal-Nr.            Fortlaufende, phys. Nummer des Kanals
	 * Byte ..          Zeile n: FG                             Funktionsgruppe des Kanals
	 * Byte ..          Zeile n: DE                             Zuzuweisende DE
	 * </pre>
	 * Inhalt Byte: Anzahl Tabellenzeilen:<br> Gibt die Länge der Konfiguration/Soll-Tabelle in Zeilen an. Die Gesamtlänge
	 * der Tabelle ergibt sich aus der Multiplikation mit der festen Zeilenlänge von 4 Bytes. Die Angabe 0 löscht die
	 * vorhandene Konfigurationstabelle.<br> <br> Inhalt Byte: E/A-Konzentrator:<br> In diesem Byte ist die OSI-2-Adresse
	 * abgelegt, über die der E/A-Konzentrator am Lokalbus adressiert wird. Ist der E/A-Konzentrator intern im SM
	 * erreichbar, so wird in diesem Byte 0 oder ein Wert größer 31 (herstellerabhängig) eingetragen.<br> <br> Inhalt Byte:
	 * E/A-Kanal-Nr.:<br> Die interne, phys. E/A-Kanalnummer im E/A-Konzentrator.<br> <br> Inhalt Byte: FG:<br> Die FG gibt
	 * die Funktionsgruppe an, zu der der Kanal gehört.<br> <br> Inhalt Byte: DE:<br> Der DE gibt die logische Kanalnummer
	 * an, mit welcher die Adressierung dieses Kanals vorgenommen werden soll.<br>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ034Id003AtgTlsSysKonfigurationsTabelleAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Zuordnungsblöcke aus Telegramm ------------------------------------------------------------------------------
		int count = deBlock.getDeData().readUnsignedByteAsInt();

		int _OSI2Adresse_;
		int _Kanal_;
		int _FG_;
		int _DE_;

		data.getArray("KonfigurationsEintrag").setLength(count);
		for (int i = 0; i < count; i++) {
			_OSI2Adresse_ = deBlock.getDeData().readUnsignedByteAsInt();
			_Kanal_ = deBlock.getDeData().readUnsignedByteAsInt();
			_FG_ = deBlock.getDeData().readUnsignedByteAsInt();
			_DE_ = deBlock.getDeData().readUnsignedByteAsInt();
			data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("OSI2Adresse").set(_OSI2Adresse_);
			data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("Kanal").set(_Kanal_);
			data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("FG").set(_FG_);
			data.getArray("KonfigurationsEintrag").getItem(i).getUnscaledValue("DE").set(_DE_);
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 254 : Typ: 35 Id: 3 Name: OSI3-Routingfeld (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 = Systemsteuerung des SM ]
	 * Byte 3           Typ der DE-Daten                        [ 35 ]
	 * Byte 4           Anzahl Routing-Blöcke                   [1..n]
	 * Byte 5           Länge Routing-Block                     [2..16]
	 * Byte 6           FG                                      [255]
	 * Byte 7           Adressidentifier                        Anzahl Adressfelder 0..7 (siehe unten)
	 * Byte 8           Adresse I, 1. Verm.-Abschnitt
	 * Byte 9           Adresse II, 1. Verm.-Abschnitt
	 * Byte ..          Adresse I, n. Verm.-Abschnitt
	 * Byte ..          Adresse II, n. Verm.-Abschnitt
	 * Byte ..          Länge Routing-Block                     [2..16]
	 * Byte ..          FG                                      [1..254]
	 * Byte ..          Adressidentifier                        Anzahl Adressfelder 0..7 (siehe unten)
	 * Byte ..          Adresse I, 1. Verm.-Abschnitt
	 * Byte ..          Adresse II, 1. Verm.-Abschnitt
	 * Byte ..          Adresse I, n. Verm.-Abschnitt
	 * Byte ..          Adresse II, n. Verm.-Abschnitt
	 * </pre>
	 * Inhalt Byte: Adressidentifier<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *    0        0                 Länge             0        0        0
	 * </pre>
	 * Länge=Anzahl der folgenden Adresspaare [ 0..7 ], bei Länge 0 folgen keine Adresspaare.<br> <br> Inhalt Byte: Adresse
	 * I und II<br> Siehe Beschreibung der Vermittlungsschicht OSI 3.<br> Die Adressen im OSI-3-Routing-Feld werden in der
	 * Reihenfolge im DE-Block aufgeführt, wie sie für eine Vermittlung von der Streckenstation in Antwortrichtung benötigt
	 * werden (d.h. das 1. Adresspaar ist Adresse SS - Adresse UZ).<br>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg254Typ035Id003AtgTlsSysOSI3RoutingFeldAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _FG_;
		int _Adresse1_;
		int _Adresse2_;
		int length;// Anzahl der Vermittlungsabschnitte (je 2 Byte) !!

		// Umsetzung
		Data data = result.getData();// Datensatz holen
		TlsData deApplicationDataTmp = new TlsData(300);// erstmal zu groß anlegen...

		int count = data.getArray("OSI3RoutingBlock").getLength();
		deApplicationDataTmp.writeIntAsByte(count);// Anzahl der Routingblöcke schreiben

		for (int i = 0; i < count; i++) {
			length = data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").getLength();
			deApplicationDataTmp.writeIntAsByte((length * 2) + 2);// Länge Routingblock schreiben

			_FG_ = data.getArray("OSI3RoutingBlock").getItem(i).getUnscaledValue("FG").intValue();
			deApplicationDataTmp.writeIntAsByte(_FG_);// Funktionsgruppe je Block schreiben

			// Die Länge der Vermittlungsabschnitte wird bitcodiert im Adressidentifier übergeben
			int adressIdentifier = (length << 3);
			deApplicationDataTmp.writeIntAsByte(adressIdentifier);// je Block Anzahl der Vermittlungsabschnitte schreiben

			for (int j = 0; j < length; j++) {
				_Adresse1_ = data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").getItem(j).getUnscaledValue("Adresse1").intValue();
				_Adresse2_ = data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").getItem(j).getUnscaledValue("Adresse2").intValue();
				deApplicationDataTmp.writeIntAsByte(_Adresse1_);// Adresspaar schreiben
				deApplicationDataTmp.writeIntAsByte(_Adresse2_);
			}
		}

		// DeBlock erzeugen
		int lengthDeBlock = deApplicationDataTmp.currentWritePosition();// tatsächliche verwendete Größe ermitteln
		TlsData deApplicationData = new TlsData(lengthDeBlock);// neuen passenden DeBlock anlagen
		deApplicationDataTmp.rewindReadPosition();
		deApplicationData = deApplicationDataTmp.readNextTlsDataBlock(lengthDeBlock);
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 35, 3, _de.getDeChannel());
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;

		//
		//		int _FG_;
		//		int _Adresse1_;
		//		int _Adresse2_;
		//
		//		int length; // Anzahl der Vermittlungsabschnitte (je 2 Byte) !!
		//
		//		int count = deBlock.getDeData().readUnsignedByteAsInt(); // Anzahl der Routingblöcke
		//		data.getArray("OSI3RoutingBlock").setLength(count);
		//
		//		for (int i = 0; i < count; i++) {
		//			deBlock.getDeData().readUnsignedByteAsInt(); // Länge Routingblock überlesen, wird nicht benötigt.
		//
		//			_FG_ = deBlock.getDeData().readUnsignedByteAsInt();
		//			data.getArray("OSI3RoutingBlock").getItem(i).getUnscaledValue("FG").set(_FG_);
		//
		//			length = deBlock.getDeData().readUnsignedByteAsInt(); // Anzahl der Vermittlungsabschnitte (je 2 Byte) !!
		//			data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").setLength(length);
		//
		//			for (int j = 0; j < length; j++) {
		//				_Adresse1_ = deBlock.getDeData().readUnsignedByteAsInt();
		//				_Adresse2_ = deBlock.getDeData().readUnsignedByteAsInt();
		//				data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").getItem(j).getUnscaledValue("Adresse1").set(_Adresse1_);
		//				data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").getItem(j).getUnscaledValue("Adresse2").set(_Adresse2_);
		//			}
		//		}
	}

	/**
	 * FG 254 : Typ: 35 Id: 3 Name: OSI3-Routingfeld (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 = Systemsteuerung des SM ]
	 * Byte 3           Typ der DE-Daten                        [ 35 ]
	 * Byte 4           Anzahl Routing-Blöcke                   [1..n]
	 * Byte 5           Länge Routing-Block                     [2..16]
	 * Byte 6           FG                                      [255]
	 * Byte 7           Adressidentifier                        Anzahl Adressfelder 0..7 (siehe unten)
	 * Byte 8           Adresse I, 1. Verm.-Abschnitt
	 * Byte 9           Adresse II, 1. Verm.-Abschnitt
	 * Byte ..          Adresse I, n. Verm.-Abschnitt
	 * Byte ..          Adresse II, n. Verm.-Abschnitt
	 * Byte ..          Länge Routing-Block                     [2..16]
	 * Byte ..          FG                                      [1..254]
	 * Byte ..          Adressidentifier                        Anzahl Adressfelder 0..7 (siehe unten)
	 * Byte ..          Adresse I, 1. Verm.-Abschnitt
	 * Byte ..          Adresse II, 1. Verm.-Abschnitt
	 * Byte ..          Adresse I, n. Verm.-Abschnitt
	 * Byte ..          Adresse II, n. Verm.-Abschnitt
	 * </pre>
	 * Inhalt Byte: Adressidentifier<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *    0        0                 Länge             0        0        0
	 * </pre>
	 * Länge=Anzahl der folgenden Adresspaare [ 0..7 ], bei Länge 0 folgen keine Adresspaare.<br> <br> Inhalt Byte: Adresse
	 * I und II<br> Siehe Beschreibung der Vermittlungsschicht OSI 3.<br> Die Adressen im OSI-3-Routing-Feld werden in der
	 * Reihenfolge im DE-Block aufgeführt, wie sie für eine Vermittlung von der Streckenstation in Antwortrichtung benötigt
	 * werden (d.h. das 1. Adresspaar ist Adresse SS - Adresse UZ).<br>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ035Id003AtgTlsSysOSI3RoutingFeldAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Routing-Blöcke aus Telegramm --------------------------------------------------------------------------------
		int _FG_;
		int _Adresse1_;
		int _Adresse2_;

		int length;// Anzahl der Vermittlungsabschnitte (je 2 Byte) !!

		int count = deBlock.getDeData().readUnsignedByteAsInt();// Anzahl der Routingblöcke
		data.getArray("OSI3RoutingBlock").setLength(count);

		for (int i = 0; i < count; i++) {
			deBlock.getDeData().readUnsignedByteAsInt();// Länge Routingblock überlesen, wird nicht benötigt.

			_FG_ = deBlock.getDeData().readUnsignedByteAsInt();
			data.getArray("OSI3RoutingBlock").getItem(i).getUnscaledValue("FG").set(_FG_);

			// Die Länge der Vermittlungsabschnitte wird bitcodiert im Adressidentifier übergeben
			int adressIdentifier = deBlock.getDeData().readUnsignedByteAsInt();
			length = ((adressIdentifier >> 3) & 0x07);
			data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").setLength(length);

			for (int j = 0; j < length; j++) {
				_Adresse1_ = deBlock.getDeData().readUnsignedByteAsInt();
				_Adresse2_ = deBlock.getDeData().readUnsignedByteAsInt();
				data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").getItem(j).getUnscaledValue("Adresse1").set(_Adresse1_);
				data.getArray("OSI3RoutingBlock").getItem(i).getArray("VermittlungsAbschnitt").getItem(j).getUnscaledValue("Adresse2").set(_Adresse2_);
			}
		}

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 254 : Typ: 37 Id: 3  Name: Knotennummer  (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          [ 5 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 = Systemsteuerung des SM]
	 * Byte 3           Typ der DE-Daten                        [ 37 ]
	 * Byte 4           Knotennummer low Byte                   [ 0..255 ]
	 * Byte 5           Knotennummer high Byte                  [ 0..255 ]
	 * Byte 6           Knotennummer extra high Byte            [ 0..255 ]
	 * </pre>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg254Typ037Id003AtgTlsSysKnotenNummerAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _KnotenNummer_;

		// Umsetzung
		Data data = result.getData();// Datensatz holen

		_KnotenNummer_ = data.getUnscaledValue("KnotenNummer").intValue();

		// DeBlock als TlsData aufbauen...
		TlsData deApplicationData = new TlsData(3);
		deApplicationData.writeIntAsLowHighExtraHighByte(_KnotenNummer_);

		// DeBlock erzeugen
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 37, 3, _de.getDeChannel());
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}

	/**
	 * FG 254 : Typ: 37 Id: 3  Name: Knotennummer  (Abruf / Antwort). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          [ 5 ]
	 * Byte 2           Daten-Endgeräte-Kanal                   [ 0 = Systemsteuerung des SM]
	 * Byte 3           Typ der DE-Daten                        [ 37 ]
	 * Byte 4           Knotennummer low Byte                   [ 0..255 ]
	 * Byte 5           Knotennummer high Byte                  [ 0..255 ]
	 * Byte 6           Knotennummer extra high Byte            [ 0..255 ]
	 * </pre>
	 *
	 * @param deBlock         Der zu bearbeitende De-Block.
	 * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
	 * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
	 *
	 * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
	 */
	public ResultData Fg254Typ037Id003AtgTlsSysKnotenNummerAspTlsZustand(DeBlock deBlock, Data data, DataDescription dataDescription) {
		ResultData resultData = null;

		// Knotennummer aus Telegramm ----------------------------------------------------------------------------------
		int _KnotenNummer_ = deBlock.getDeData().readUnsignedLowHighExtraHighByteAsInt();
		data.getUnscaledValue("KnotenNummer").set(_KnotenNummer_);

		//Ergebnisdatensatz erzeugen...
		resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
		return resultData;
	}

	/**
	 * FG 254 : Typ: 38 Id: 2  Name: Reset  (Abruf). TLS-Datenbeschreibung:<br>
	 * <pre>
	 * Position         Bezeichnung                             Erläuterung
	 * Byte 1           Länge DE-Block                          Länge des folgenden DE-Blocks [ 3 ]
	 * Byte 2           Daten-Endgeräte-Kanal (DE)              [0..254, 255]
	 * Byte 3           Typ der DE-Daten                        [ 38 ]
	 * Byte 4           Grund für Reset                         siehe unten
	 * </pre>
	 * Inhalt Byte: Grund für Reset:<br>
	 * <pre>
	 *    7        6        5        4        3        2        1        0
	 *        Reserviert für EAK               Reserviert für Steuermodul
	 *
	 * Bit 0:            0:         keine                                 Reaktion
	 *                   1:         vollständiger Wiederanlauf des Steuermoduls (Hardwarereset)
	 * Bit 1:          = 0:         reserviert
	 * Bit 2:          = 0:         reserviert
	 * Bit 3:          = 0:         reserviert
	 * Bit 4:            0:         keine Reaktion
	 *                   1:         vollständiger Wiederanlauf des EAKs (Hardwarereset)
	 * Bit 5:          = 0:         reserviert
	 * Bit 6:          = 0:         reserviert
	 * Bit 7:          = 0:         reserviert
	 * </pre>
	 * Der Grund für den Reset kann getrennt für Steuermodul und E/A-Konzentratoren eingestellt werden. Sollte es nicht
	 * möglich sein (z.B. bei internen EAKs), die E/A-Konzentratoren unabhängig vom Steuermodul zu behandeln, so wird immer
	 * der Grund vom Steuermodul im E/A-Konzentrator übernommen.<br>
	 *
	 * @param result Der vom Datenverteiler empgangene Datensatz, der in einen De-Block konvertiert wird.
	 *
	 * @return De-Block mit den konvertierten Daten.
	 */
	public DeBlock Fg254Typ038Id002AtgTlsSysResetAspTlsVorgabe(ResultData result) {
		DeBlock deBlock = null;

		// Datenblock beseht aus folgenden Inhalten:
		int _SM_;
		int _EAK_;

		// Umsetzung
		Data data = result.getData();// Datensatz holen

		_SM_ = data.getUnscaledValue("SM").intValue();
		_EAK_ = data.getUnscaledValue("EAK").intValue();

		// DeBlock als TlsData aufbauen...
		TlsData deApplicationData = new TlsData(1);
		deApplicationData.writeIntAsByte((_EAK_ << 4) + _SM_);

		// DeBlock erzeugen
		try {
			deBlock = new DeBlock(deApplicationData, _de.getEak().getSm().getOsi7Adress(), _de.getFg(), 38, 2, _de.getDeChannel());
		}
		catch (IllegalOsi7TelegramStructureException e) {
			debug.warning("Fehler beim Erzeugen eines De-Blocks zum Versand: " + e);
		}
		return deBlock;
	}
}
